/* This file is automatically generated from cssproperties.in by makeprop, do not edit */
/* Copyright 1998 W. Bastian */

#ifndef CSSPROPERTIES_H
#define CSSPROPERTIES_H

DOM::DOMString getPropertyName(unsigned short id) KHTML_NO_EXPORT;

#define CSS_PROP_INVALID 0
#define CSS_PROP_MIN 1
#define CSS_PROP_BACKGROUND_COLOR 1
#define CSS_PROP_BACKGROUND_IMAGE 2
#define CSS_PROP_BACKGROUND_REPEAT 3
#define CSS_PROP_BACKGROUND_ATTACHMENT 4
#define CSS_PROP_BACKGROUND_POSITION 5
#define CSS_PROP_BACKGROUND_POSITION_X 6
#define CSS_PROP_BACKGROUND_POSITION_Y 7
#define CSS_PROP_BACKGROUND_CLIP 8
#define CSS_PROP_BACKGROUND_ORIGIN 9
#define CSS_PROP_BACKGROUND_SIZE 10
#define CSS_PROP_BORDER_TOP_RIGHT_RADIUS 11
#define CSS_PROP_BORDER_BOTTOM_RIGHT_RADIUS 12
#define CSS_PROP_BORDER_BOTTOM_LEFT_RADIUS 13
#define CSS_PROP_BORDER_TOP_LEFT_RADIUS 14
#define CSS_PROP_BORDER_RADIUS 15
#define CSS_PROP_BORDER_COLLAPSE 16
#define CSS_PROP_BORDER_SPACING 17
#define CSS_PROP__KHTML_BORDER_HORIZONTAL_SPACING 18
#define CSS_PROP__KHTML_BORDER_VERTICAL_SPACING 19
#define CSS_PROP_BORDER_TOP_COLOR 20
#define CSS_PROP_BORDER_RIGHT_COLOR 21
#define CSS_PROP_BORDER_BOTTOM_COLOR 22
#define CSS_PROP_BORDER_LEFT_COLOR 23
#define CSS_PROP_BORDER_TOP_STYLE 24
#define CSS_PROP_BORDER_RIGHT_STYLE 25
#define CSS_PROP_BORDER_BOTTOM_STYLE 26
#define CSS_PROP_BORDER_LEFT_STYLE 27
#define CSS_PROP_BORDER_TOP_WIDTH 28
#define CSS_PROP_BORDER_RIGHT_WIDTH 29
#define CSS_PROP_BORDER_BOTTOM_WIDTH 30
#define CSS_PROP_BORDER_LEFT_WIDTH 31
#define CSS_PROP_BOTTOM 32
#define CSS_PROP_CAPTION_SIDE 33
#define CSS_PROP_CLEAR 34
#define CSS_PROP_CLIP 35
#define CSS_PROP_COLOR 36
#define CSS_PROP_CONTENT 37
#define CSS_PROP_COUNTER_INCREMENT 38
#define CSS_PROP_COUNTER_RESET 39
#define CSS_PROP_CURSOR 40
#define CSS_PROP_DIRECTION 41
#define CSS_PROP_DISPLAY 42
#define CSS_PROP_EMPTY_CELLS 43
#define CSS_PROP_FLOAT 44
#define CSS_PROP_FONT_FAMILY 45
#define CSS_PROP_FONT_SIZE 46
#define CSS_PROP_FONT_STYLE 47
#define CSS_PROP_FONT_VARIANT 48
#define CSS_PROP_FONT_WEIGHT 49
#define CSS_PROP_HEIGHT 50
#define CSS_PROP_LEFT 51
#define CSS_PROP_LETTER_SPACING 52
#define CSS_PROP_LINE_HEIGHT 53
#define CSS_PROP_LIST_STYLE_IMAGE 54
#define CSS_PROP_LIST_STYLE_POSITION 55
#define CSS_PROP_LIST_STYLE_TYPE 56
#define CSS_PROP_MARGIN_TOP 57
#define CSS_PROP_MARGIN_RIGHT 58
#define CSS_PROP_MARGIN_BOTTOM 59
#define CSS_PROP_MARGIN_LEFT 60
#define CSS_PROP__KHTML_MARGIN_START 61
#define CSS_PROP__KHTML_MARQUEE 62
#define CSS_PROP__KHTML_MARQUEE_DIRECTION 63
#define CSS_PROP__KHTML_MARQUEE_INCREMENT 64
#define CSS_PROP__KHTML_MARQUEE_REPETITION 65
#define CSS_PROP__KHTML_MARQUEE_SPEED 66
#define CSS_PROP__KHTML_MARQUEE_STYLE 67
#define CSS_PROP_MAX_HEIGHT 68
#define CSS_PROP_MAX_WIDTH 69
#define CSS_PROP_MIN_HEIGHT 70
#define CSS_PROP_MIN_WIDTH 71
#define CSS_PROP_ORPHANS 72
#define CSS_PROP_OPACITY 73
#define CSS_PROP_OUTLINE_COLOR 74
#define CSS_PROP_OUTLINE_OFFSET 75
#define CSS_PROP_OUTLINE_STYLE 76
#define CSS_PROP_OUTLINE_WIDTH 77
#define CSS_PROP_OVERFLOW 78
#define CSS_PROP_OVERFLOW_X 79
#define CSS_PROP_OVERFLOW_Y 80
#define CSS_PROP_PADDING_TOP 81
#define CSS_PROP_PADDING_RIGHT 82
#define CSS_PROP_PADDING_BOTTOM 83
#define CSS_PROP_PADDING_LEFT 84
#define CSS_PROP__KHTML_PADDING_START 85
#define CSS_PROP_PAGE_BREAK_AFTER 86
#define CSS_PROP_PAGE_BREAK_BEFORE 87
#define CSS_PROP_PAGE_BREAK_INSIDE 88
#define CSS_PROP_POSITION 89
#define CSS_PROP_QUOTES 90
#define CSS_PROP_RIGHT 91
#define CSS_PROP_SIZE 92
#define CSS_PROP_SRC 93
#define CSS_PROP_TABLE_LAYOUT 94
#define CSS_PROP_TEXT_ALIGN 95
#define CSS_PROP_TEXT_DECORATION 96
#define CSS_PROP_TEXT_INDENT 97
#define CSS_PROP_TEXT_OVERFLOW 98
#define CSS_PROP_TEXT_SHADOW 99
#define CSS_PROP_TEXT_TRANSFORM 100
#define CSS_PROP_TOP 101
#define CSS_PROP_UNICODE_BIDI 102
#define CSS_PROP_UNICODE_RANGE 103
#define CSS_PROP_VERTICAL_ALIGN 104
#define CSS_PROP_VISIBILITY 105
#define CSS_PROP_WHITE_SPACE 106
#define CSS_PROP_WIDOWS 107
#define CSS_PROP_WIDTH 108
#define CSS_PROP_WORD_WRAP 109
#define CSS_PROP_WORD_SPACING 110
#define CSS_PROP_Z_INDEX 111
#define CSS_PROP_BACKGROUND 112
#define CSS_PROP_BORDER 113
#define CSS_PROP_BORDER_COLOR 114
#define CSS_PROP_BORDER_STYLE 115
#define CSS_PROP_BORDER_TOP 116
#define CSS_PROP_BORDER_RIGHT 117
#define CSS_PROP_BORDER_BOTTOM 118
#define CSS_PROP_BORDER_LEFT 119
#define CSS_PROP_BORDER_WIDTH 120
#define CSS_PROP_BOX_SIZING 121
#define CSS_PROP_FONT 122
#define CSS_PROP_LIST_STYLE 123
#define CSS_PROP_MARGIN 124
#define CSS_PROP_OUTLINE 125
#define CSS_PROP_PADDING 126
#define CSS_PROP_SCROLLBAR_BASE_COLOR 127
#define CSS_PROP_SCROLLBAR_FACE_COLOR 128
#define CSS_PROP_SCROLLBAR_SHADOW_COLOR 129
#define CSS_PROP_SCROLLBAR_HIGHLIGHT_COLOR 130
#define CSS_PROP_SCROLLBAR_3DLIGHT_COLOR 131
#define CSS_PROP_SCROLLBAR_DARKSHADOW_COLOR 132
#define CSS_PROP_SCROLLBAR_TRACK_COLOR 133
#define CSS_PROP_SCROLLBAR_ARROW_COLOR 134
#define CSS_PROP__KHTML_FLOW_MODE 135
#define CSS_PROP__KHTML_USER_INPUT 136
#define CSS_PROP_ALIGNMENT_BASELINE 137
#define CSS_PROP_BASELINE_SHIFT 138
#define CSS_PROP_CLIP_PATH 139
#define CSS_PROP_CLIP_RULE 140
#define CSS_PROP_COLOR_INTERPOLATION 141
#define CSS_PROP_COLOR_INTERPOLATION_FILTERS 142
#define CSS_PROP_COLOR_PROFILE 143
#define CSS_PROP_COLOR_RENDERING 144
#define CSS_PROP_DOMINANT_BASELINE 145
#define CSS_PROP_ENABLE_BACKGROUND 146
#define CSS_PROP_FILL 147
#define CSS_PROP_FILL_OPACITY 148
#define CSS_PROP_FILL_RULE 149
#define CSS_PROP_FILTER 150
#define CSS_PROP_FLOOD_COLOR 151
#define CSS_PROP_FLOOD_OPACITY 152
#define CSS_PROP_FONT_SIZE_ADJUST 153
#define CSS_PROP_FONT_STRETCH 154
#define CSS_PROP_GLYPH_ORIENTATION_HORIZONTAL 155
#define CSS_PROP_GLYPH_ORIENTATION_VERTICAL 156
#define CSS_PROP_IMAGE_RENDERING 157
#define CSS_PROP_KERNING 158
#define CSS_PROP_LIGHTING_COLOR 159
#define CSS_PROP_MARKER 160
#define CSS_PROP_MARKER_END 161
#define CSS_PROP_MARKER_MID 162
#define CSS_PROP_MARKER_START 163
#define CSS_PROP_MASK 164
#define CSS_PROP_POINTER_EVENTS 165
#define CSS_PROP_SHAPE_RENDERING 166
#define CSS_PROP_STOP_COLOR 167
#define CSS_PROP_STOP_OPACITY 168
#define CSS_PROP_STROKE 169
#define CSS_PROP_STROKE_DASHARRAY 170
#define CSS_PROP_STROKE_DASHOFFSET 171
#define CSS_PROP_STROKE_LINECAP 172
#define CSS_PROP_STROKE_LINEJOIN 173
#define CSS_PROP_STROKE_MITERLIMIT 174
#define CSS_PROP_STROKE_OPACITY 175
#define CSS_PROP_STROKE_WIDTH 176
#define CSS_PROP_TEXT_ANCHOR 177
#define CSS_PROP_TEXT_RENDERING 178
#define CSS_PROP_WRITING_MODE 179

#define CSS_PROP_TOTAL 180
const size_t maxCSSPropertyNameLength = 32;
#endif

