#ifndef SVGNAMES_H
#define SVGNAMES_H

#include "misc/htmlnames.h"
#include "dom/QualifiedName.h"
#include "xml/Document.h"

#define idAttr ATTR_ID

namespace WebCore
{
namespace SVGNames
{
void init();
extern DOM::QualifiedName aTag;
extern DOM::QualifiedName altGlyphTag;
extern DOM::QualifiedName altGlyphDefTag;
extern DOM::QualifiedName altGlyphItemTag;
extern DOM::QualifiedName animateTag;
extern DOM::QualifiedName animateColorTag;
extern DOM::QualifiedName animateMotionTag;
extern DOM::QualifiedName animateTransformTag;
extern DOM::QualifiedName setTag;
extern DOM::QualifiedName circleTag;
extern DOM::QualifiedName clipPathTag;
extern DOM::QualifiedName color_profileTag;
extern DOM::QualifiedName cursorTag;
extern DOM::QualifiedName definition_srcTag;
extern DOM::QualifiedName defsTag;
extern DOM::QualifiedName descTag;
extern DOM::QualifiedName ellipseTag;
extern DOM::QualifiedName feBlendTag;
extern DOM::QualifiedName feColorMatrixTag;
extern DOM::QualifiedName feComponentTransferTag;
extern DOM::QualifiedName feCompositeTag;
extern DOM::QualifiedName feConvolveMatrixTag;
extern DOM::QualifiedName feDiffuseLightingTag;
extern DOM::QualifiedName feDisplacementMapTag;
extern DOM::QualifiedName feDistantLightTag;
extern DOM::QualifiedName feFloodTag;
extern DOM::QualifiedName feFuncATag;
extern DOM::QualifiedName feFuncBTag;
extern DOM::QualifiedName feFuncGTag;
extern DOM::QualifiedName feFuncRTag;
extern DOM::QualifiedName feGaussianBlurTag;
extern DOM::QualifiedName feImageTag;
extern DOM::QualifiedName feMergeTag;
extern DOM::QualifiedName feMergeNodeTag;
extern DOM::QualifiedName feMorphologyTag;
extern DOM::QualifiedName feOffsetTag;
extern DOM::QualifiedName fePointLightTag;
extern DOM::QualifiedName feSpecularLightingTag;
extern DOM::QualifiedName feSpotLightTag;
extern DOM::QualifiedName feTileTag;
extern DOM::QualifiedName feTurbulenceTag;
extern DOM::QualifiedName filterTag;
extern DOM::QualifiedName fontTag;
extern DOM::QualifiedName font_faceTag;
extern DOM::QualifiedName font_face_formatTag;
extern DOM::QualifiedName font_face_nameTag;
extern DOM::QualifiedName font_face_srcTag;
extern DOM::QualifiedName font_face_uriTag;
extern DOM::QualifiedName foreignObjectTag;
extern DOM::QualifiedName gTag;
extern DOM::QualifiedName glyphTag;
extern DOM::QualifiedName glyphRefTag;
extern DOM::QualifiedName hkernTag;
extern DOM::QualifiedName imageTag;
extern DOM::QualifiedName lineTag;
extern DOM::QualifiedName linearGradientTag;
extern DOM::QualifiedName markerTag;
extern DOM::QualifiedName maskTag;
extern DOM::QualifiedName metadataTag;
extern DOM::QualifiedName missing_glyphTag;
extern DOM::QualifiedName mpathTag;
extern DOM::QualifiedName pathTag;
extern DOM::QualifiedName patternTag;
extern DOM::QualifiedName polygonTag;
extern DOM::QualifiedName polylineTag;
extern DOM::QualifiedName radialGradientTag;
extern DOM::QualifiedName rectTag;
extern DOM::QualifiedName scriptTag;
extern DOM::QualifiedName stopTag;
extern DOM::QualifiedName styleTag;
extern DOM::QualifiedName svgTag;
extern DOM::QualifiedName switchTag;
extern DOM::QualifiedName symbolTag;
extern DOM::QualifiedName textTag;
extern DOM::QualifiedName textPathTag;
extern DOM::QualifiedName titleTag;
extern DOM::QualifiedName trefTag;
extern DOM::QualifiedName tspanTag;
extern DOM::QualifiedName useTag;
extern DOM::QualifiedName viewTag;
extern DOM::QualifiedName vkernTag;
extern DOM::QualifiedName accent_heightAttr;
extern DOM::QualifiedName accumulateAttr;
extern DOM::QualifiedName additiveAttr;
extern DOM::QualifiedName alignment_baselineAttr;
extern DOM::QualifiedName alphabeticAttr;
extern DOM::QualifiedName amplitudeAttr;
extern DOM::QualifiedName animateAttr;
extern DOM::QualifiedName arabic_formAttr;
extern DOM::QualifiedName ascentAttr;
extern DOM::QualifiedName attributeNameAttr;
extern DOM::QualifiedName attributeTypeAttr;
extern DOM::QualifiedName azimuthAttr;
extern DOM::QualifiedName baseFrequencyAttr;
extern DOM::QualifiedName baseline_shiftAttr;
extern DOM::QualifiedName baseProfileAttr;
extern DOM::QualifiedName bboxAttr;
extern DOM::QualifiedName beginAttr;
extern DOM::QualifiedName biasAttr;
extern DOM::QualifiedName byAttr;
extern DOM::QualifiedName calcModeAttr;
extern DOM::QualifiedName cap_heightAttr;
extern DOM::QualifiedName clipAttr;
extern DOM::QualifiedName clip_pathAttr;
extern DOM::QualifiedName clip_ruleAttr;
extern DOM::QualifiedName clipPathUnitsAttr;
extern DOM::QualifiedName colorAttr;
extern DOM::QualifiedName color_interpolationAttr;
extern DOM::QualifiedName color_interpolation_filtersAttr;
extern DOM::QualifiedName color_profileAttr;
extern DOM::QualifiedName color_renderingAttr;
extern DOM::QualifiedName contentScriptTypeAttr;
extern DOM::QualifiedName contentStyleTypeAttr;
extern DOM::QualifiedName cursorAttr;
extern DOM::QualifiedName cxAttr;
extern DOM::QualifiedName cyAttr;
extern DOM::QualifiedName dAttr;
extern DOM::QualifiedName descentAttr;
extern DOM::QualifiedName diffuseConstantAttr;
extern DOM::QualifiedName directionAttr;
extern DOM::QualifiedName displayAttr;
extern DOM::QualifiedName divisorAttr;
extern DOM::QualifiedName dominant_baselineAttr;
extern DOM::QualifiedName durAttr;
extern DOM::QualifiedName dxAttr;
extern DOM::QualifiedName dyAttr;
extern DOM::QualifiedName edgeModeAttr;
extern DOM::QualifiedName elevationAttr;
extern DOM::QualifiedName enable_backgroundAttr;
extern DOM::QualifiedName endAttr;
extern DOM::QualifiedName exponentAttr;
extern DOM::QualifiedName externalResourcesRequiredAttr;
extern DOM::QualifiedName feColorMatrixAttr;
extern DOM::QualifiedName feCompositeAttr;
extern DOM::QualifiedName feGaussianBlurAttr;
extern DOM::QualifiedName feMorphologyAttr;
extern DOM::QualifiedName feTileAttr;
extern DOM::QualifiedName fillAttr;
extern DOM::QualifiedName fill_opacityAttr;
extern DOM::QualifiedName fill_ruleAttr;
extern DOM::QualifiedName filterAttr;
extern DOM::QualifiedName filterResAttr;
extern DOM::QualifiedName filterUnitsAttr;
extern DOM::QualifiedName flood_colorAttr;
extern DOM::QualifiedName flood_opacityAttr;
extern DOM::QualifiedName font_familyAttr;
extern DOM::QualifiedName font_sizeAttr;
extern DOM::QualifiedName font_size_adjustAttr;
extern DOM::QualifiedName font_stretchAttr;
extern DOM::QualifiedName font_styleAttr;
extern DOM::QualifiedName font_variantAttr;
extern DOM::QualifiedName font_weightAttr;
extern DOM::QualifiedName formatAttr;
extern DOM::QualifiedName fromAttr;
extern DOM::QualifiedName fxAttr;
extern DOM::QualifiedName fyAttr;
extern DOM::QualifiedName g1Attr;
extern DOM::QualifiedName g2Attr;
extern DOM::QualifiedName glyph_nameAttr;
extern DOM::QualifiedName glyph_orientation_horizontalAttr;
extern DOM::QualifiedName glyph_orientation_verticalAttr;
extern DOM::QualifiedName glyphRefAttr;
extern DOM::QualifiedName gradientTransformAttr;
extern DOM::QualifiedName gradientUnitsAttr;
extern DOM::QualifiedName hangingAttr;
extern DOM::QualifiedName heightAttr;
extern DOM::QualifiedName horiz_adv_xAttr;
extern DOM::QualifiedName horiz_origin_xAttr;
extern DOM::QualifiedName horiz_origin_yAttr;
extern DOM::QualifiedName ideographicAttr;
extern DOM::QualifiedName image_renderingAttr;
extern DOM::QualifiedName inAttr;
extern DOM::QualifiedName in2Attr;
extern DOM::QualifiedName interceptAttr;
extern DOM::QualifiedName kAttr;
extern DOM::QualifiedName k1Attr;
extern DOM::QualifiedName k2Attr;
extern DOM::QualifiedName k3Attr;
extern DOM::QualifiedName k4Attr;
extern DOM::QualifiedName kernelMatrixAttr;
extern DOM::QualifiedName kernelUnitLengthAttr;
extern DOM::QualifiedName kerningAttr;
extern DOM::QualifiedName keyPointsAttr;
extern DOM::QualifiedName keySplinesAttr;
extern DOM::QualifiedName keyTimesAttr;
extern DOM::QualifiedName langAttr;
extern DOM::QualifiedName lengthAdjustAttr;
extern DOM::QualifiedName letter_spacingAttr;
extern DOM::QualifiedName lighting_colorAttr;
extern DOM::QualifiedName limitingConeAngleAttr;
extern DOM::QualifiedName localAttr;
extern DOM::QualifiedName marker_endAttr;
extern DOM::QualifiedName marker_midAttr;
extern DOM::QualifiedName marker_startAttr;
extern DOM::QualifiedName markerHeightAttr;
extern DOM::QualifiedName markerUnitsAttr;
extern DOM::QualifiedName markerWidthAttr;
extern DOM::QualifiedName maskAttr;
extern DOM::QualifiedName maskContentUnitsAttr;
extern DOM::QualifiedName maskUnitsAttr;
extern DOM::QualifiedName mathematicalAttr;
extern DOM::QualifiedName maxAttr;
extern DOM::QualifiedName mediaAttr;
extern DOM::QualifiedName methodAttr;
extern DOM::QualifiedName minAttr;
extern DOM::QualifiedName modeAttr;
extern DOM::QualifiedName nameAttr;
extern DOM::QualifiedName numOctavesAttr;
extern DOM::QualifiedName offsetAttr;
extern DOM::QualifiedName onactivateAttr;
extern DOM::QualifiedName onbeginAttr;
extern DOM::QualifiedName onendAttr;
extern DOM::QualifiedName onfocusinAttr;
extern DOM::QualifiedName onfocusoutAttr;
extern DOM::QualifiedName onrepeatAttr;
extern DOM::QualifiedName onzoomAttr;
extern DOM::QualifiedName opacityAttr;
extern DOM::QualifiedName operatorAttr;
extern DOM::QualifiedName orderAttr;
extern DOM::QualifiedName orientAttr;
extern DOM::QualifiedName orientationAttr;
extern DOM::QualifiedName originAttr;
extern DOM::QualifiedName overflowAttr;
extern DOM::QualifiedName overline_positionAttr;
extern DOM::QualifiedName overline_thicknessAttr;
extern DOM::QualifiedName panose_1Attr;
extern DOM::QualifiedName pathAttr;
extern DOM::QualifiedName pathLengthAttr;
extern DOM::QualifiedName patternContentUnitsAttr;
extern DOM::QualifiedName patternTransformAttr;
extern DOM::QualifiedName patternUnitsAttr;
extern DOM::QualifiedName pointer_eventsAttr;
extern DOM::QualifiedName pointsAttr;
extern DOM::QualifiedName pointsAtXAttr;
extern DOM::QualifiedName pointsAtYAttr;
extern DOM::QualifiedName pointsAtZAttr;
extern DOM::QualifiedName preserveAlphaAttr;
extern DOM::QualifiedName preserveAspectRatioAttr;
extern DOM::QualifiedName primitiveUnitsAttr;
extern DOM::QualifiedName rAttr;
extern DOM::QualifiedName radiusAttr;
extern DOM::QualifiedName refXAttr;
extern DOM::QualifiedName refYAttr;
extern DOM::QualifiedName rendering_intentAttr;
extern DOM::QualifiedName repeatCountAttr;
extern DOM::QualifiedName repeatDurAttr;
extern DOM::QualifiedName requiredExtensionsAttr;
extern DOM::QualifiedName requiredFeaturesAttr;
extern DOM::QualifiedName restartAttr;
extern DOM::QualifiedName resultAttr;
extern DOM::QualifiedName rotateAttr;
extern DOM::QualifiedName rxAttr;
extern DOM::QualifiedName ryAttr;
extern DOM::QualifiedName scaleAttr;
extern DOM::QualifiedName seedAttr;
extern DOM::QualifiedName shape_renderingAttr;
extern DOM::QualifiedName slopeAttr;
extern DOM::QualifiedName spacingAttr;
extern DOM::QualifiedName specularConstantAttr;
extern DOM::QualifiedName specularExponentAttr;
extern DOM::QualifiedName spreadMethodAttr;
extern DOM::QualifiedName startOffsetAttr;
extern DOM::QualifiedName stdDeviationAttr;
extern DOM::QualifiedName stemhAttr;
extern DOM::QualifiedName stemvAttr;
extern DOM::QualifiedName stitchTilesAttr;
extern DOM::QualifiedName stop_colorAttr;
extern DOM::QualifiedName stop_opacityAttr;
extern DOM::QualifiedName strikethrough_positionAttr;
extern DOM::QualifiedName strikethrough_thicknessAttr;
extern DOM::QualifiedName strokeAttr;
extern DOM::QualifiedName stroke_dasharrayAttr;
extern DOM::QualifiedName stroke_dashoffsetAttr;
extern DOM::QualifiedName stroke_linecapAttr;
extern DOM::QualifiedName stroke_linejoinAttr;
extern DOM::QualifiedName stroke_miterlimitAttr;
extern DOM::QualifiedName stroke_opacityAttr;
extern DOM::QualifiedName stroke_widthAttr;
extern DOM::QualifiedName styleAttr;
extern DOM::QualifiedName surfaceScaleAttr;
extern DOM::QualifiedName systemLanguageAttr;
extern DOM::QualifiedName tableValuesAttr;
extern DOM::QualifiedName targetAttr;
extern DOM::QualifiedName targetXAttr;
extern DOM::QualifiedName targetYAttr;
extern DOM::QualifiedName text_anchorAttr;
extern DOM::QualifiedName text_decorationAttr;
extern DOM::QualifiedName text_renderingAttr;
extern DOM::QualifiedName textLengthAttr;
extern DOM::QualifiedName titleAttr;
extern DOM::QualifiedName toAttr;
extern DOM::QualifiedName transformAttr;
extern DOM::QualifiedName typeAttr;
extern DOM::QualifiedName u1Attr;
extern DOM::QualifiedName u2Attr;
extern DOM::QualifiedName underline_positionAttr;
extern DOM::QualifiedName underline_thicknessAttr;
extern DOM::QualifiedName unicodeAttr;
extern DOM::QualifiedName unicode_bidiAttr;
extern DOM::QualifiedName unicode_rangeAttr;
extern DOM::QualifiedName units_per_emAttr;
extern DOM::QualifiedName v_alphabeticAttr;
extern DOM::QualifiedName v_hangingAttr;
extern DOM::QualifiedName v_ideographicAttr;
extern DOM::QualifiedName v_mathematicalAttr;
extern DOM::QualifiedName valuesAttr;
extern DOM::QualifiedName versionAttr;
extern DOM::QualifiedName vert_adv_yAttr;
extern DOM::QualifiedName vert_origin_xAttr;
extern DOM::QualifiedName vert_origin_yAttr;
extern DOM::QualifiedName viewBoxAttr;
extern DOM::QualifiedName viewTargetAttr;
extern DOM::QualifiedName visibilityAttr;
extern DOM::QualifiedName widthAttr;
extern DOM::QualifiedName widthsAttr;
extern DOM::QualifiedName word_spacingAttr;
extern DOM::QualifiedName writing_modeAttr;
extern DOM::QualifiedName xAttr;
extern DOM::QualifiedName x_heightAttr;
extern DOM::QualifiedName x1Attr;
extern DOM::QualifiedName x2Attr;
extern DOM::QualifiedName xChannelSelectorAttr;
extern DOM::QualifiedName yAttr;
extern DOM::QualifiedName y1Attr;
extern DOM::QualifiedName y2Attr;
extern DOM::QualifiedName yChannelSelectorAttr;
extern DOM::QualifiedName zAttr;
extern DOM::QualifiedName zoomAndPanAttr;
}
namespace HTMLNames
{
extern DOM::QualifiedName classAttr;
}
}

#endif
