/*
 * XPathNamespaceImpl.h - Copyright 2005 Frerich Raabe <raabe@kde.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef XPATHNAMESPACEIMPL_H
#define XPATHNAMESPACEIMPL_H

#include <xml/dom_nodeimpl.h>

namespace DOM
{
class DocumentImpl;
class DOMStringImpl;

class XPathNamespaceImpl : public NodeImpl
{
public:
    XPathNamespaceImpl(DocumentImpl *ptr,
                       DOMStringImpl *prefix, DOMStringImpl *uri);
    XPathNamespaceImpl(const XPathNamespaceImpl &other);
    XPathNamespaceImpl &operator=(const XPathNamespaceImpl &rhs);
    virtual ~XPathNamespaceImpl();

    virtual DocumentImpl *ownerDocument() const;
    virtual ElementImpl *ownerElement() const;

    virtual const AtomicString &prefix() const;
    virtual DOMString nodeName() const;
    virtual const AtomicString &namespaceURI() const;

    virtual unsigned short nodeType() const;

private:
    DocumentImpl *m_ownerDocument;
    AtomicString m_prefix;
    AtomicString m_uri;
};

}

#endif // XPATHNAMESPACEIMPL_H

