/*
    Copyright (C) 2004, 2005, 2007, 2008 Nikolas Zimmermann <zimmermann@kde.org>
                  2004, 2005, 2006 Rob Buis <buis@kde.org>

    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include "wtf/Platform.h"

#if ENABLE(SVG)
#include "SVGGElement.h"

#include "RenderSVGTransformableContainer.h"

namespace WebCore
{

SVGGElement::SVGGElement(const QualifiedName &tagName, Document *doc)
    : SVGStyledTransformableElement(tagName, doc)
    , SVGTests()
    , SVGLangSpace()
    , SVGExternalResourcesRequired()
{
}

SVGGElement::~SVGGElement()
{
}

void SVGGElement::parseMappedAttribute(MappedAttribute *attr)
{
    if (SVGTests::parseMappedAttribute(attr)) {
        return;
    }
    if (SVGLangSpace::parseMappedAttribute(attr)) {
        return;
    }
    if (SVGExternalResourcesRequired::parseMappedAttribute(attr)) {
        return;
    }

    SVGStyledTransformableElement::parseMappedAttribute(attr);
}

void SVGGElement::svgAttributeChanged(const QualifiedName &attrName)
{
    SVGStyledTransformableElement::svgAttributeChanged(attrName);

    if (!renderer()) {
        return;
    }

    if (SVGTests::isKnownAttribute(attrName) ||
            SVGLangSpace::isKnownAttribute(attrName) ||
            SVGExternalResourcesRequired::isKnownAttribute(attrName) ||
            SVGStyledTransformableElement::isKnownAttribute(attrName)) {
        renderer()->setNeedsLayout(true);
    }
}

void SVGGElement::childrenChanged(bool changedByParser, Node *beforeChange, Node *afterChange, int childCountDelta)
{
    SVGStyledTransformableElement::childrenChanged(changedByParser, beforeChange, afterChange, childCountDelta);

    if (renderer()) {
        renderer()->setNeedsLayout(true);
    }
}

RenderObject *SVGGElement::createRenderer(RenderArena *arena, RenderStyle *)
{
    return new(arena) RenderSVGTransformableContainer(this);
}

}

#endif // ENABLE(SVG)
