#include "svg/SVGNames.h"

#include "svg/XLinkNames.h"

namespace WebCore
{
namespace HTMLNames
{
DOM::QualifiedName classAttr;

}
namespace XLinkNames
{
DOM::QualifiedName actuateAttr;
DOM::QualifiedName arcroleAttr;
DOM::QualifiedName hrefAttr;
DOM::QualifiedName roleAttr;
DOM::QualifiedName showAttr;
DOM::QualifiedName titleAttr;
DOM::QualifiedName typeAttr;

}
namespace SVGNames
{
DOM::QualifiedName aTag;
DOM::QualifiedName altGlyphTag;
DOM::QualifiedName altGlyphDefTag;
DOM::QualifiedName altGlyphItemTag;
DOM::QualifiedName animateTag;
DOM::QualifiedName animateColorTag;
DOM::QualifiedName animateMotionTag;
DOM::QualifiedName animateTransformTag;
DOM::QualifiedName setTag;
DOM::QualifiedName circleTag;
DOM::QualifiedName clipPathTag;
DOM::QualifiedName color_profileTag;
DOM::QualifiedName cursorTag;
DOM::QualifiedName definition_srcTag;
DOM::QualifiedName defsTag;
DOM::QualifiedName descTag;
DOM::QualifiedName ellipseTag;
DOM::QualifiedName feBlendTag;
DOM::QualifiedName feColorMatrixTag;
DOM::QualifiedName feComponentTransferTag;
DOM::QualifiedName feCompositeTag;
DOM::QualifiedName feConvolveMatrixTag;
DOM::QualifiedName feDiffuseLightingTag;
DOM::QualifiedName feDisplacementMapTag;
DOM::QualifiedName feDistantLightTag;
DOM::QualifiedName feFloodTag;
DOM::QualifiedName feFuncATag;
DOM::QualifiedName feFuncBTag;
DOM::QualifiedName feFuncGTag;
DOM::QualifiedName feFuncRTag;
DOM::QualifiedName feGaussianBlurTag;
DOM::QualifiedName feImageTag;
DOM::QualifiedName feMergeTag;
DOM::QualifiedName feMergeNodeTag;
DOM::QualifiedName feMorphologyTag;
DOM::QualifiedName feOffsetTag;
DOM::QualifiedName fePointLightTag;
DOM::QualifiedName feSpecularLightingTag;
DOM::QualifiedName feSpotLightTag;
DOM::QualifiedName feTileTag;
DOM::QualifiedName feTurbulenceTag;
DOM::QualifiedName filterTag;
DOM::QualifiedName fontTag;
DOM::QualifiedName font_faceTag;
DOM::QualifiedName font_face_formatTag;
DOM::QualifiedName font_face_nameTag;
DOM::QualifiedName font_face_srcTag;
DOM::QualifiedName font_face_uriTag;
DOM::QualifiedName foreignObjectTag;
DOM::QualifiedName gTag;
DOM::QualifiedName glyphTag;
DOM::QualifiedName glyphRefTag;
DOM::QualifiedName hkernTag;
DOM::QualifiedName imageTag;
DOM::QualifiedName lineTag;
DOM::QualifiedName linearGradientTag;
DOM::QualifiedName markerTag;
DOM::QualifiedName maskTag;
DOM::QualifiedName metadataTag;
DOM::QualifiedName missing_glyphTag;
DOM::QualifiedName mpathTag;
DOM::QualifiedName pathTag;
DOM::QualifiedName patternTag;
DOM::QualifiedName polygonTag;
DOM::QualifiedName polylineTag;
DOM::QualifiedName radialGradientTag;
DOM::QualifiedName rectTag;
DOM::QualifiedName scriptTag;
DOM::QualifiedName stopTag;
DOM::QualifiedName styleTag;
DOM::QualifiedName svgTag;
DOM::QualifiedName switchTag;
DOM::QualifiedName symbolTag;
DOM::QualifiedName textTag;
DOM::QualifiedName textPathTag;
DOM::QualifiedName titleTag;
DOM::QualifiedName trefTag;
DOM::QualifiedName tspanTag;
DOM::QualifiedName useTag;
DOM::QualifiedName viewTag;
DOM::QualifiedName vkernTag;
DOM::QualifiedName accent_heightAttr;
DOM::QualifiedName accumulateAttr;
DOM::QualifiedName additiveAttr;
DOM::QualifiedName alignment_baselineAttr;
DOM::QualifiedName alphabeticAttr;
DOM::QualifiedName amplitudeAttr;
DOM::QualifiedName animateAttr;
DOM::QualifiedName arabic_formAttr;
DOM::QualifiedName ascentAttr;
DOM::QualifiedName attributeNameAttr;
DOM::QualifiedName attributeTypeAttr;
DOM::QualifiedName azimuthAttr;
DOM::QualifiedName baseFrequencyAttr;
DOM::QualifiedName baseline_shiftAttr;
DOM::QualifiedName baseProfileAttr;
DOM::QualifiedName bboxAttr;
DOM::QualifiedName beginAttr;
DOM::QualifiedName biasAttr;
DOM::QualifiedName byAttr;
DOM::QualifiedName calcModeAttr;
DOM::QualifiedName cap_heightAttr;
DOM::QualifiedName clipAttr;
DOM::QualifiedName clip_pathAttr;
DOM::QualifiedName clip_ruleAttr;
DOM::QualifiedName clipPathUnitsAttr;
DOM::QualifiedName colorAttr;
DOM::QualifiedName color_interpolationAttr;
DOM::QualifiedName color_interpolation_filtersAttr;
DOM::QualifiedName color_profileAttr;
DOM::QualifiedName color_renderingAttr;
DOM::QualifiedName contentScriptTypeAttr;
DOM::QualifiedName contentStyleTypeAttr;
DOM::QualifiedName cursorAttr;
DOM::QualifiedName cxAttr;
DOM::QualifiedName cyAttr;
DOM::QualifiedName dAttr;
DOM::QualifiedName descentAttr;
DOM::QualifiedName diffuseConstantAttr;
DOM::QualifiedName directionAttr;
DOM::QualifiedName displayAttr;
DOM::QualifiedName divisorAttr;
DOM::QualifiedName dominant_baselineAttr;
DOM::QualifiedName durAttr;
DOM::QualifiedName dxAttr;
DOM::QualifiedName dyAttr;
DOM::QualifiedName edgeModeAttr;
DOM::QualifiedName elevationAttr;
DOM::QualifiedName enable_backgroundAttr;
DOM::QualifiedName endAttr;
DOM::QualifiedName exponentAttr;
DOM::QualifiedName externalResourcesRequiredAttr;
DOM::QualifiedName feColorMatrixAttr;
DOM::QualifiedName feCompositeAttr;
DOM::QualifiedName feGaussianBlurAttr;
DOM::QualifiedName feMorphologyAttr;
DOM::QualifiedName feTileAttr;
DOM::QualifiedName fillAttr;
DOM::QualifiedName fill_opacityAttr;
DOM::QualifiedName fill_ruleAttr;
DOM::QualifiedName filterAttr;
DOM::QualifiedName filterResAttr;
DOM::QualifiedName filterUnitsAttr;
DOM::QualifiedName flood_colorAttr;
DOM::QualifiedName flood_opacityAttr;
DOM::QualifiedName font_familyAttr;
DOM::QualifiedName font_sizeAttr;
DOM::QualifiedName font_size_adjustAttr;
DOM::QualifiedName font_stretchAttr;
DOM::QualifiedName font_styleAttr;
DOM::QualifiedName font_variantAttr;
DOM::QualifiedName font_weightAttr;
DOM::QualifiedName formatAttr;
DOM::QualifiedName fromAttr;
DOM::QualifiedName fxAttr;
DOM::QualifiedName fyAttr;
DOM::QualifiedName g1Attr;
DOM::QualifiedName g2Attr;
DOM::QualifiedName glyph_nameAttr;
DOM::QualifiedName glyph_orientation_horizontalAttr;
DOM::QualifiedName glyph_orientation_verticalAttr;
DOM::QualifiedName glyphRefAttr;
DOM::QualifiedName gradientTransformAttr;
DOM::QualifiedName gradientUnitsAttr;
DOM::QualifiedName hangingAttr;
DOM::QualifiedName heightAttr;
DOM::QualifiedName horiz_adv_xAttr;
DOM::QualifiedName horiz_origin_xAttr;
DOM::QualifiedName horiz_origin_yAttr;
DOM::QualifiedName ideographicAttr;
DOM::QualifiedName image_renderingAttr;
DOM::QualifiedName inAttr;
DOM::QualifiedName in2Attr;
DOM::QualifiedName interceptAttr;
DOM::QualifiedName kAttr;
DOM::QualifiedName k1Attr;
DOM::QualifiedName k2Attr;
DOM::QualifiedName k3Attr;
DOM::QualifiedName k4Attr;
DOM::QualifiedName kernelMatrixAttr;
DOM::QualifiedName kernelUnitLengthAttr;
DOM::QualifiedName kerningAttr;
DOM::QualifiedName keyPointsAttr;
DOM::QualifiedName keySplinesAttr;
DOM::QualifiedName keyTimesAttr;
DOM::QualifiedName langAttr;
DOM::QualifiedName lengthAdjustAttr;
DOM::QualifiedName letter_spacingAttr;
DOM::QualifiedName lighting_colorAttr;
DOM::QualifiedName limitingConeAngleAttr;
DOM::QualifiedName localAttr;
DOM::QualifiedName marker_endAttr;
DOM::QualifiedName marker_midAttr;
DOM::QualifiedName marker_startAttr;
DOM::QualifiedName markerHeightAttr;
DOM::QualifiedName markerUnitsAttr;
DOM::QualifiedName markerWidthAttr;
DOM::QualifiedName maskAttr;
DOM::QualifiedName maskContentUnitsAttr;
DOM::QualifiedName maskUnitsAttr;
DOM::QualifiedName mathematicalAttr;
DOM::QualifiedName maxAttr;
DOM::QualifiedName mediaAttr;
DOM::QualifiedName methodAttr;
DOM::QualifiedName minAttr;
DOM::QualifiedName modeAttr;
DOM::QualifiedName nameAttr;
DOM::QualifiedName numOctavesAttr;
DOM::QualifiedName offsetAttr;
DOM::QualifiedName onactivateAttr;
DOM::QualifiedName onbeginAttr;
DOM::QualifiedName onendAttr;
DOM::QualifiedName onfocusinAttr;
DOM::QualifiedName onfocusoutAttr;
DOM::QualifiedName onrepeatAttr;
DOM::QualifiedName onzoomAttr;
DOM::QualifiedName opacityAttr;
DOM::QualifiedName operatorAttr;
DOM::QualifiedName orderAttr;
DOM::QualifiedName orientAttr;
DOM::QualifiedName orientationAttr;
DOM::QualifiedName originAttr;
DOM::QualifiedName overflowAttr;
DOM::QualifiedName overline_positionAttr;
DOM::QualifiedName overline_thicknessAttr;
DOM::QualifiedName panose_1Attr;
DOM::QualifiedName pathAttr;
DOM::QualifiedName pathLengthAttr;
DOM::QualifiedName patternContentUnitsAttr;
DOM::QualifiedName patternTransformAttr;
DOM::QualifiedName patternUnitsAttr;
DOM::QualifiedName pointer_eventsAttr;
DOM::QualifiedName pointsAttr;
DOM::QualifiedName pointsAtXAttr;
DOM::QualifiedName pointsAtYAttr;
DOM::QualifiedName pointsAtZAttr;
DOM::QualifiedName preserveAlphaAttr;
DOM::QualifiedName preserveAspectRatioAttr;
DOM::QualifiedName primitiveUnitsAttr;
DOM::QualifiedName rAttr;
DOM::QualifiedName radiusAttr;
DOM::QualifiedName refXAttr;
DOM::QualifiedName refYAttr;
DOM::QualifiedName rendering_intentAttr;
DOM::QualifiedName repeatCountAttr;
DOM::QualifiedName repeatDurAttr;
DOM::QualifiedName requiredExtensionsAttr;
DOM::QualifiedName requiredFeaturesAttr;
DOM::QualifiedName restartAttr;
DOM::QualifiedName resultAttr;
DOM::QualifiedName rotateAttr;
DOM::QualifiedName rxAttr;
DOM::QualifiedName ryAttr;
DOM::QualifiedName scaleAttr;
DOM::QualifiedName seedAttr;
DOM::QualifiedName shape_renderingAttr;
DOM::QualifiedName slopeAttr;
DOM::QualifiedName spacingAttr;
DOM::QualifiedName specularConstantAttr;
DOM::QualifiedName specularExponentAttr;
DOM::QualifiedName spreadMethodAttr;
DOM::QualifiedName startOffsetAttr;
DOM::QualifiedName stdDeviationAttr;
DOM::QualifiedName stemhAttr;
DOM::QualifiedName stemvAttr;
DOM::QualifiedName stitchTilesAttr;
DOM::QualifiedName stop_colorAttr;
DOM::QualifiedName stop_opacityAttr;
DOM::QualifiedName strikethrough_positionAttr;
DOM::QualifiedName strikethrough_thicknessAttr;
DOM::QualifiedName strokeAttr;
DOM::QualifiedName stroke_dasharrayAttr;
DOM::QualifiedName stroke_dashoffsetAttr;
DOM::QualifiedName stroke_linecapAttr;
DOM::QualifiedName stroke_linejoinAttr;
DOM::QualifiedName stroke_miterlimitAttr;
DOM::QualifiedName stroke_opacityAttr;
DOM::QualifiedName stroke_widthAttr;
DOM::QualifiedName styleAttr;
DOM::QualifiedName surfaceScaleAttr;
DOM::QualifiedName systemLanguageAttr;
DOM::QualifiedName tableValuesAttr;
DOM::QualifiedName targetAttr;
DOM::QualifiedName targetXAttr;
DOM::QualifiedName targetYAttr;
DOM::QualifiedName text_anchorAttr;
DOM::QualifiedName text_decorationAttr;
DOM::QualifiedName text_renderingAttr;
DOM::QualifiedName textLengthAttr;
DOM::QualifiedName titleAttr;
DOM::QualifiedName toAttr;
DOM::QualifiedName transformAttr;
DOM::QualifiedName typeAttr;
DOM::QualifiedName u1Attr;
DOM::QualifiedName u2Attr;
DOM::QualifiedName underline_positionAttr;
DOM::QualifiedName underline_thicknessAttr;
DOM::QualifiedName unicodeAttr;
DOM::QualifiedName unicode_bidiAttr;
DOM::QualifiedName unicode_rangeAttr;
DOM::QualifiedName units_per_emAttr;
DOM::QualifiedName v_alphabeticAttr;
DOM::QualifiedName v_hangingAttr;
DOM::QualifiedName v_ideographicAttr;
DOM::QualifiedName v_mathematicalAttr;
DOM::QualifiedName valuesAttr;
DOM::QualifiedName versionAttr;
DOM::QualifiedName vert_adv_yAttr;
DOM::QualifiedName vert_origin_xAttr;
DOM::QualifiedName vert_origin_yAttr;
DOM::QualifiedName viewBoxAttr;
DOM::QualifiedName viewTargetAttr;
DOM::QualifiedName visibilityAttr;
DOM::QualifiedName widthAttr;
DOM::QualifiedName widthsAttr;
DOM::QualifiedName word_spacingAttr;
DOM::QualifiedName writing_modeAttr;
DOM::QualifiedName xAttr;
DOM::QualifiedName x_heightAttr;
DOM::QualifiedName x1Attr;
DOM::QualifiedName x2Attr;
DOM::QualifiedName xChannelSelectorAttr;
DOM::QualifiedName yAttr;
DOM::QualifiedName y1Attr;
DOM::QualifiedName y2Attr;
DOM::QualifiedName yChannelSelectorAttr;
DOM::QualifiedName zAttr;
DOM::QualifiedName zoomAndPanAttr;

void init()
{
    aTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_A), DOM::emptyPrefixName);
    altGlyphTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ALTGLYPH), DOM::emptyPrefixName);
    altGlyphDefTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ALTGLYPHDEF), DOM::emptyPrefixName);
    altGlyphItemTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ALTGLYPHITEM), DOM::emptyPrefixName);
    animateTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ANIMATE), DOM::emptyPrefixName);
    animateColorTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ANIMATECOLOR), DOM::emptyPrefixName);
    animateMotionTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ANIMATEMOTION), DOM::emptyPrefixName);
    animateTransformTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ANIMATETRANSFORM), DOM::emptyPrefixName);
    setTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_SET), DOM::emptyPrefixName);
    circleTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_CIRCLE), DOM::emptyPrefixName);
    clipPathTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_CLIPPATH), DOM::emptyPrefixName);
    color_profileTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_COLOR_PROFILE), DOM::emptyPrefixName);
    cursorTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_CURSOR), DOM::emptyPrefixName);
    definition_srcTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_DEFINITION_SRC), DOM::emptyPrefixName);
    defsTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_DEFS), DOM::emptyPrefixName);
    descTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_DESC), DOM::emptyPrefixName);
    ellipseTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_ELLIPSE), DOM::emptyPrefixName);
    feBlendTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEBLEND), DOM::emptyPrefixName);
    feColorMatrixTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FECOLORMATRIX), DOM::emptyPrefixName);
    feComponentTransferTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FECOMPONENTTRANSFER), DOM::emptyPrefixName);
    feCompositeTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FECOMPOSITE), DOM::emptyPrefixName);
    feConvolveMatrixTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FECONVOLVEMATRIX), DOM::emptyPrefixName);
    feDiffuseLightingTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEDIFFUSELIGHTING), DOM::emptyPrefixName);
    feDisplacementMapTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEDISPLACEMENTMAP), DOM::emptyPrefixName);
    feDistantLightTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEDISTANTLIGHT), DOM::emptyPrefixName);
    feFloodTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEFLOOD), DOM::emptyPrefixName);
    feFuncATag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEFUNCA), DOM::emptyPrefixName);
    feFuncBTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEFUNCB), DOM::emptyPrefixName);
    feFuncGTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEFUNCG), DOM::emptyPrefixName);
    feFuncRTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEFUNCR), DOM::emptyPrefixName);
    feGaussianBlurTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEGAUSSIANBLUR), DOM::emptyPrefixName);
    feImageTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEIMAGE), DOM::emptyPrefixName);
    feMergeTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEMERGE), DOM::emptyPrefixName);
    feMergeNodeTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEMERGENODE), DOM::emptyPrefixName);
    feMorphologyTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEMORPHOLOGY), DOM::emptyPrefixName);
    feOffsetTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEOFFSET), DOM::emptyPrefixName);
    fePointLightTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FEPOINTLIGHT), DOM::emptyPrefixName);
    feSpecularLightingTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FESPECULARLIGHTING), DOM::emptyPrefixName);
    feSpotLightTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FESPOTLIGHT), DOM::emptyPrefixName);
    feTileTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FETILE), DOM::emptyPrefixName);
    feTurbulenceTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FETURBULENCE), DOM::emptyPrefixName);
    filterTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FILTER), DOM::emptyPrefixName);
    fontTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FONT), DOM::emptyPrefixName);
    font_faceTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FONT_FACE), DOM::emptyPrefixName);
    font_face_formatTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FONT_FACE_FORMAT), DOM::emptyPrefixName);
    font_face_nameTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FONT_FACE_NAME), DOM::emptyPrefixName);
    font_face_srcTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FONT_FACE_SRC), DOM::emptyPrefixName);
    font_face_uriTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FONT_FACE_URI), DOM::emptyPrefixName);
    foreignObjectTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_FOREIGNOBJECT), DOM::emptyPrefixName);
    gTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_G), DOM::emptyPrefixName);
    glyphTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_GLYPH), DOM::emptyPrefixName);
    glyphRefTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_GLYPHREF), DOM::emptyPrefixName);
    hkernTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_HKERN), DOM::emptyPrefixName);
    imageTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_IMAGE), DOM::emptyPrefixName);
    lineTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_LINE), DOM::emptyPrefixName);
    linearGradientTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_LINEARGRADIENT), DOM::emptyPrefixName);
    markerTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_MARKER), DOM::emptyPrefixName);
    maskTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_MASK), DOM::emptyPrefixName);
    metadataTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_METADATA), DOM::emptyPrefixName);
    missing_glyphTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_MISSING_GLYPH), DOM::emptyPrefixName);
    mpathTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_MPATH), DOM::emptyPrefixName);
    pathTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_PATH), DOM::emptyPrefixName);
    patternTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_PATTERN), DOM::emptyPrefixName);
    polygonTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_POLYGON), DOM::emptyPrefixName);
    polylineTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_POLYLINE), DOM::emptyPrefixName);
    radialGradientTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_RADIALGRADIENT), DOM::emptyPrefixName);
    rectTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_RECT), DOM::emptyPrefixName);
    scriptTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_SCRIPT), DOM::emptyPrefixName);
    stopTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_STOP), DOM::emptyPrefixName);
    styleTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_STYLE), DOM::emptyPrefixName);
    svgTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_SVG), DOM::emptyPrefixName);
    switchTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_SWITCH), DOM::emptyPrefixName);
    symbolTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_SYMBOL), DOM::emptyPrefixName);
    textTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, DOM::localNamePart(ATTR_TEXT)), DOM::emptyPrefixName);
    textPathTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_TEXTPATH), DOM::emptyPrefixName);
    titleTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_TITLE), DOM::emptyPrefixName);
    trefTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_TREF), DOM::emptyPrefixName);
    tspanTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_TSPAN), DOM::emptyPrefixName);
    useTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_USE), DOM::emptyPrefixName);
    viewTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_VIEW), DOM::emptyPrefixName);
    vkernTag = DOM::QualifiedName(DOM::makeId(DOM::svgNamespace, ID_VKERN), DOM::emptyPrefixName);
    accent_heightAttr = DOM::QualifiedName(ATTR_ACCENT_HEIGHT, DOM::emptyPrefixName);
    accumulateAttr = DOM::QualifiedName(ATTR_ACCUMULATE, DOM::emptyPrefixName);
    additiveAttr = DOM::QualifiedName(ATTR_ADDITIVE, DOM::emptyPrefixName);
    alignment_baselineAttr = DOM::QualifiedName(ATTR_ALIGNMENT_BASELINE, DOM::emptyPrefixName);
    alphabeticAttr = DOM::QualifiedName(ATTR_ALPHABETIC, DOM::emptyPrefixName);
    amplitudeAttr = DOM::QualifiedName(ATTR_AMPLITUDE, DOM::emptyPrefixName);
    animateAttr = DOM::QualifiedName(ATTR_ANIMATE, DOM::emptyPrefixName);
    arabic_formAttr = DOM::QualifiedName(ATTR_ARABIC_FORM, DOM::emptyPrefixName);
    ascentAttr = DOM::QualifiedName(ATTR_ASCENT, DOM::emptyPrefixName);
    attributeNameAttr = DOM::QualifiedName(ATTR_ATTRIBUTENAME, DOM::emptyPrefixName);
    attributeTypeAttr = DOM::QualifiedName(ATTR_ATTRIBUTETYPE, DOM::emptyPrefixName);
    azimuthAttr = DOM::QualifiedName(ATTR_AZIMUTH, DOM::emptyPrefixName);
    baseFrequencyAttr = DOM::QualifiedName(ATTR_BASEFREQUENCY, DOM::emptyPrefixName);
    baseline_shiftAttr = DOM::QualifiedName(ATTR_BASELINE_SHIFT, DOM::emptyPrefixName);
    baseProfileAttr = DOM::QualifiedName(ATTR_BASEPROFILE, DOM::emptyPrefixName);
    bboxAttr = DOM::QualifiedName(ATTR_BBOX, DOM::emptyPrefixName);
    beginAttr = DOM::QualifiedName(ATTR_BEGIN, DOM::emptyPrefixName);
    biasAttr = DOM::QualifiedName(ATTR_BIAS, DOM::emptyPrefixName);
    byAttr = DOM::QualifiedName(ATTR_BY, DOM::emptyPrefixName);
    calcModeAttr = DOM::QualifiedName(ATTR_CALCMODE, DOM::emptyPrefixName);
    cap_heightAttr = DOM::QualifiedName(ATTR_CAP_HEIGHT, DOM::emptyPrefixName);
    clipAttr = DOM::QualifiedName(ATTR_CLIP, DOM::emptyPrefixName);
    clip_pathAttr = DOM::QualifiedName(ATTR_CLIP_PATH, DOM::emptyPrefixName);
    clip_ruleAttr = DOM::QualifiedName(ATTR_CLIP_RULE, DOM::emptyPrefixName);
    clipPathUnitsAttr = DOM::QualifiedName(ATTR_CLIPPATHUNITS, DOM::emptyPrefixName);
    colorAttr = DOM::QualifiedName(ATTR_COLOR, DOM::emptyPrefixName);
    color_interpolationAttr = DOM::QualifiedName(ATTR_COLOR_INTERPOLATION, DOM::emptyPrefixName);
    color_interpolation_filtersAttr = DOM::QualifiedName(ATTR_COLOR_INTERPOLATION_FILTERS, DOM::emptyPrefixName);
    color_profileAttr = DOM::QualifiedName(ATTR_COLOR_PROFILE, DOM::emptyPrefixName);
    color_renderingAttr = DOM::QualifiedName(ATTR_COLOR_RENDERING, DOM::emptyPrefixName);
    contentScriptTypeAttr = DOM::QualifiedName(ATTR_CONTENTSCRIPTTYPE, DOM::emptyPrefixName);
    contentStyleTypeAttr = DOM::QualifiedName(ATTR_CONTENTSTYLETYPE, DOM::emptyPrefixName);
    cursorAttr = DOM::QualifiedName(ATTR_CURSOR, DOM::emptyPrefixName);
    cxAttr = DOM::QualifiedName(ATTR_CX, DOM::emptyPrefixName);
    cyAttr = DOM::QualifiedName(ATTR_CY, DOM::emptyPrefixName);
    dAttr = DOM::QualifiedName(ATTR_D, DOM::emptyPrefixName);
    descentAttr = DOM::QualifiedName(ATTR_DESCENT, DOM::emptyPrefixName);
    diffuseConstantAttr = DOM::QualifiedName(ATTR_DIFFUSECONSTANT, DOM::emptyPrefixName);
    directionAttr = DOM::QualifiedName(ATTR_DIRECTION, DOM::emptyPrefixName);
    displayAttr = DOM::QualifiedName(ATTR_DISPLAY, DOM::emptyPrefixName);
    divisorAttr = DOM::QualifiedName(ATTR_DIVISOR, DOM::emptyPrefixName);
    dominant_baselineAttr = DOM::QualifiedName(ATTR_DOMINANT_BASELINE, DOM::emptyPrefixName);
    durAttr = DOM::QualifiedName(ATTR_DUR, DOM::emptyPrefixName);
    dxAttr = DOM::QualifiedName(ATTR_DX, DOM::emptyPrefixName);
    dyAttr = DOM::QualifiedName(ATTR_DY, DOM::emptyPrefixName);
    edgeModeAttr = DOM::QualifiedName(ATTR_EDGEMODE, DOM::emptyPrefixName);
    elevationAttr = DOM::QualifiedName(ATTR_ELEVATION, DOM::emptyPrefixName);
    enable_backgroundAttr = DOM::QualifiedName(ATTR_ENABLE_BACKGROUND, DOM::emptyPrefixName);
    endAttr = DOM::QualifiedName(ATTR_END, DOM::emptyPrefixName);
    exponentAttr = DOM::QualifiedName(ATTR_EXPONENT, DOM::emptyPrefixName);
    externalResourcesRequiredAttr = DOM::QualifiedName(ATTR_EXTERNALRESOURCESREQUIRED, DOM::emptyPrefixName);
    feColorMatrixAttr = DOM::QualifiedName(ATTR_FECOLORMATRIX, DOM::emptyPrefixName);
    feCompositeAttr = DOM::QualifiedName(ATTR_FECOMPOSITE, DOM::emptyPrefixName);
    feGaussianBlurAttr = DOM::QualifiedName(ATTR_FEGAUSSIANBLUR, DOM::emptyPrefixName);
    feMorphologyAttr = DOM::QualifiedName(ATTR_FEMORPHOLOGY, DOM::emptyPrefixName);
    feTileAttr = DOM::QualifiedName(ATTR_FETILE, DOM::emptyPrefixName);
    fillAttr = DOM::QualifiedName(ATTR_FILL, DOM::emptyPrefixName);
    fill_opacityAttr = DOM::QualifiedName(ATTR_FILL_OPACITY, DOM::emptyPrefixName);
    fill_ruleAttr = DOM::QualifiedName(ATTR_FILL_RULE, DOM::emptyPrefixName);
    filterAttr = DOM::QualifiedName(ATTR_FILTER, DOM::emptyPrefixName);
    filterResAttr = DOM::QualifiedName(ATTR_FILTERRES, DOM::emptyPrefixName);
    filterUnitsAttr = DOM::QualifiedName(ATTR_FILTERUNITS, DOM::emptyPrefixName);
    flood_colorAttr = DOM::QualifiedName(ATTR_FLOOD_COLOR, DOM::emptyPrefixName);
    flood_opacityAttr = DOM::QualifiedName(ATTR_FLOOD_OPACITY, DOM::emptyPrefixName);
    font_familyAttr = DOM::QualifiedName(ATTR_FONT_FAMILY, DOM::emptyPrefixName);
    font_sizeAttr = DOM::QualifiedName(ATTR_FONT_SIZE, DOM::emptyPrefixName);
    font_size_adjustAttr = DOM::QualifiedName(ATTR_FONT_SIZE_ADJUST, DOM::emptyPrefixName);
    font_stretchAttr = DOM::QualifiedName(ATTR_FONT_STRETCH, DOM::emptyPrefixName);
    font_styleAttr = DOM::QualifiedName(ATTR_FONT_STYLE, DOM::emptyPrefixName);
    font_variantAttr = DOM::QualifiedName(ATTR_FONT_VARIANT, DOM::emptyPrefixName);
    font_weightAttr = DOM::QualifiedName(ATTR_FONT_WEIGHT, DOM::emptyPrefixName);
    formatAttr = DOM::QualifiedName(ATTR_FORMAT, DOM::emptyPrefixName);
    fromAttr = DOM::QualifiedName(ATTR_FROM, DOM::emptyPrefixName);
    fxAttr = DOM::QualifiedName(ATTR_FX, DOM::emptyPrefixName);
    fyAttr = DOM::QualifiedName(ATTR_FY, DOM::emptyPrefixName);
    g1Attr = DOM::QualifiedName(ATTR_G1, DOM::emptyPrefixName);
    g2Attr = DOM::QualifiedName(ATTR_G2, DOM::emptyPrefixName);
    glyph_nameAttr = DOM::QualifiedName(ATTR_GLYPH_NAME, DOM::emptyPrefixName);
    glyph_orientation_horizontalAttr = DOM::QualifiedName(ATTR_GLYPH_ORIENTATION_HORIZONTAL, DOM::emptyPrefixName);
    glyph_orientation_verticalAttr = DOM::QualifiedName(ATTR_GLYPH_ORIENTATION_VERTICAL, DOM::emptyPrefixName);
    glyphRefAttr = DOM::QualifiedName(ATTR_GLYPHREF, DOM::emptyPrefixName);
    gradientTransformAttr = DOM::QualifiedName(ATTR_GRADIENTTRANSFORM, DOM::emptyPrefixName);
    gradientUnitsAttr = DOM::QualifiedName(ATTR_GRADIENTUNITS, DOM::emptyPrefixName);
    hangingAttr = DOM::QualifiedName(ATTR_HANGING, DOM::emptyPrefixName);
    heightAttr = DOM::QualifiedName(ATTR_HEIGHT, DOM::emptyPrefixName);
    horiz_adv_xAttr = DOM::QualifiedName(ATTR_HORIZ_ADV_X, DOM::emptyPrefixName);
    horiz_origin_xAttr = DOM::QualifiedName(ATTR_HORIZ_ORIGIN_X, DOM::emptyPrefixName);
    horiz_origin_yAttr = DOM::QualifiedName(ATTR_HORIZ_ORIGIN_Y, DOM::emptyPrefixName);
    ideographicAttr = DOM::QualifiedName(ATTR_IDEOGRAPHIC, DOM::emptyPrefixName);
    image_renderingAttr = DOM::QualifiedName(ATTR_IMAGE_RENDERING, DOM::emptyPrefixName);
    inAttr = DOM::QualifiedName(ATTR_IN, DOM::emptyPrefixName);
    in2Attr = DOM::QualifiedName(ATTR_IN2, DOM::emptyPrefixName);
    interceptAttr = DOM::QualifiedName(ATTR_INTERCEPT, DOM::emptyPrefixName);
    kAttr = DOM::QualifiedName(ATTR_K, DOM::emptyPrefixName);
    k1Attr = DOM::QualifiedName(ATTR_K1, DOM::emptyPrefixName);
    k2Attr = DOM::QualifiedName(ATTR_K2, DOM::emptyPrefixName);
    k3Attr = DOM::QualifiedName(ATTR_K3, DOM::emptyPrefixName);
    k4Attr = DOM::QualifiedName(ATTR_K4, DOM::emptyPrefixName);
    kernelMatrixAttr = DOM::QualifiedName(ATTR_KERNELMATRIX, DOM::emptyPrefixName);
    kernelUnitLengthAttr = DOM::QualifiedName(ATTR_KERNELUNITLENGTH, DOM::emptyPrefixName);
    kerningAttr = DOM::QualifiedName(ATTR_KERNING, DOM::emptyPrefixName);
    keyPointsAttr = DOM::QualifiedName(ATTR_KEYPOINTS, DOM::emptyPrefixName);
    keySplinesAttr = DOM::QualifiedName(ATTR_KEYSPLINES, DOM::emptyPrefixName);
    keyTimesAttr = DOM::QualifiedName(ATTR_KEYTIMES, DOM::emptyPrefixName);
    langAttr = DOM::QualifiedName(ATTR_LANG, DOM::emptyPrefixName);
    lengthAdjustAttr = DOM::QualifiedName(ATTR_LENGTHADJUST, DOM::emptyPrefixName);
    letter_spacingAttr = DOM::QualifiedName(ATTR_LETTER_SPACING, DOM::emptyPrefixName);
    lighting_colorAttr = DOM::QualifiedName(ATTR_LIGHTING_COLOR, DOM::emptyPrefixName);
    limitingConeAngleAttr = DOM::QualifiedName(ATTR_LIMITINGCONEANGLE, DOM::emptyPrefixName);
    localAttr = DOM::QualifiedName(ATTR_LOCAL, DOM::emptyPrefixName);
    marker_endAttr = DOM::QualifiedName(ATTR_MARKER_END, DOM::emptyPrefixName);
    marker_midAttr = DOM::QualifiedName(ATTR_MARKER_MID, DOM::emptyPrefixName);
    marker_startAttr = DOM::QualifiedName(ATTR_MARKER_START, DOM::emptyPrefixName);
    markerHeightAttr = DOM::QualifiedName(ATTR_MARKERHEIGHT, DOM::emptyPrefixName);
    markerUnitsAttr = DOM::QualifiedName(ATTR_MARKERUNITS, DOM::emptyPrefixName);
    markerWidthAttr = DOM::QualifiedName(ATTR_MARKERWIDTH, DOM::emptyPrefixName);
    maskAttr = DOM::QualifiedName(ATTR_MASK, DOM::emptyPrefixName);
    maskContentUnitsAttr = DOM::QualifiedName(ATTR_MASKCONTENTUNITS, DOM::emptyPrefixName);
    maskUnitsAttr = DOM::QualifiedName(ATTR_MASKUNITS, DOM::emptyPrefixName);
    mathematicalAttr = DOM::QualifiedName(ATTR_MATHEMATICAL, DOM::emptyPrefixName);
    maxAttr = DOM::QualifiedName(ATTR_MAX, DOM::emptyPrefixName);
    mediaAttr = DOM::QualifiedName(ATTR_MEDIA, DOM::emptyPrefixName);
    methodAttr = DOM::QualifiedName(ATTR_METHOD, DOM::emptyPrefixName);
    minAttr = DOM::QualifiedName(ATTR_MIN, DOM::emptyPrefixName);
    modeAttr = DOM::QualifiedName(ATTR_MODE, DOM::emptyPrefixName);
    nameAttr = DOM::QualifiedName(ATTR_NAME, DOM::emptyPrefixName);
    numOctavesAttr = DOM::QualifiedName(ATTR_NUMOCTAVES, DOM::emptyPrefixName);
    offsetAttr = DOM::QualifiedName(ATTR_OFFSET, DOM::emptyPrefixName);
    onactivateAttr = DOM::QualifiedName(ATTR_ONACTIVATE, DOM::emptyPrefixName);
    onbeginAttr = DOM::QualifiedName(ATTR_ONBEGIN, DOM::emptyPrefixName);
    onendAttr = DOM::QualifiedName(ATTR_ONEND, DOM::emptyPrefixName);
    onfocusinAttr = DOM::QualifiedName(ATTR_ONFOCUSIN, DOM::emptyPrefixName);
    onfocusoutAttr = DOM::QualifiedName(ATTR_ONFOCUSOUT, DOM::emptyPrefixName);
    onrepeatAttr = DOM::QualifiedName(ATTR_ONREPEAT, DOM::emptyPrefixName);
    onzoomAttr = DOM::QualifiedName(ATTR_ONZOOM, DOM::emptyPrefixName);
    opacityAttr = DOM::QualifiedName(ATTR_OPACITY, DOM::emptyPrefixName);
    operatorAttr = DOM::QualifiedName(ATTR_OPERATOR, DOM::emptyPrefixName);
    orderAttr = DOM::QualifiedName(ATTR_ORDER, DOM::emptyPrefixName);
    orientAttr = DOM::QualifiedName(ATTR_ORIENT, DOM::emptyPrefixName);
    orientationAttr = DOM::QualifiedName(ATTR_ORIENTATION, DOM::emptyPrefixName);
    originAttr = DOM::QualifiedName(ATTR_ORIGIN, DOM::emptyPrefixName);
    overflowAttr = DOM::QualifiedName(ATTR_OVERFLOW, DOM::emptyPrefixName);
    overline_positionAttr = DOM::QualifiedName(ATTR_OVERLINE_POSITION, DOM::emptyPrefixName);
    overline_thicknessAttr = DOM::QualifiedName(ATTR_OVERLINE_THICKNESS, DOM::emptyPrefixName);
    panose_1Attr = DOM::QualifiedName(ATTR_PANOSE_1, DOM::emptyPrefixName);
    pathAttr = DOM::QualifiedName(ATTR_PATH, DOM::emptyPrefixName);
    pathLengthAttr = DOM::QualifiedName(ATTR_PATHLENGTH, DOM::emptyPrefixName);
    patternContentUnitsAttr = DOM::QualifiedName(ATTR_PATTERNCONTENTUNITS, DOM::emptyPrefixName);
    patternTransformAttr = DOM::QualifiedName(ATTR_PATTERNTRANSFORM, DOM::emptyPrefixName);
    patternUnitsAttr = DOM::QualifiedName(ATTR_PATTERNUNITS, DOM::emptyPrefixName);
    pointer_eventsAttr = DOM::QualifiedName(ATTR_POINTER_EVENTS, DOM::emptyPrefixName);
    pointsAttr = DOM::QualifiedName(ATTR_POINTS, DOM::emptyPrefixName);
    pointsAtXAttr = DOM::QualifiedName(ATTR_POINTSATX, DOM::emptyPrefixName);
    pointsAtYAttr = DOM::QualifiedName(ATTR_POINTSATY, DOM::emptyPrefixName);
    pointsAtZAttr = DOM::QualifiedName(ATTR_POINTSATZ, DOM::emptyPrefixName);
    preserveAlphaAttr = DOM::QualifiedName(ATTR_PRESERVEALPHA, DOM::emptyPrefixName);
    preserveAspectRatioAttr = DOM::QualifiedName(ATTR_PRESERVEASPECTRATIO, DOM::emptyPrefixName);
    primitiveUnitsAttr = DOM::QualifiedName(ATTR_PRIMITIVEUNITS, DOM::emptyPrefixName);
    rAttr = DOM::QualifiedName(ATTR_R, DOM::emptyPrefixName);
    radiusAttr = DOM::QualifiedName(ATTR_RADIUS, DOM::emptyPrefixName);
    refXAttr = DOM::QualifiedName(ATTR_REFX, DOM::emptyPrefixName);
    refYAttr = DOM::QualifiedName(ATTR_REFY, DOM::emptyPrefixName);
    rendering_intentAttr = DOM::QualifiedName(ATTR_RENDERING_INTENT, DOM::emptyPrefixName);
    repeatCountAttr = DOM::QualifiedName(ATTR_REPEATCOUNT, DOM::emptyPrefixName);
    repeatDurAttr = DOM::QualifiedName(ATTR_REPEATDUR, DOM::emptyPrefixName);
    requiredExtensionsAttr = DOM::QualifiedName(ATTR_REQUIREDEXTENSIONS, DOM::emptyPrefixName);
    requiredFeaturesAttr = DOM::QualifiedName(ATTR_REQUIREDFEATURES, DOM::emptyPrefixName);
    restartAttr = DOM::QualifiedName(ATTR_RESTART, DOM::emptyPrefixName);
    resultAttr = DOM::QualifiedName(ATTR_RESULT, DOM::emptyPrefixName);
    rotateAttr = DOM::QualifiedName(ATTR_ROTATE, DOM::emptyPrefixName);
    rxAttr = DOM::QualifiedName(ATTR_RX, DOM::emptyPrefixName);
    ryAttr = DOM::QualifiedName(ATTR_RY, DOM::emptyPrefixName);
    scaleAttr = DOM::QualifiedName(ATTR_SCALE, DOM::emptyPrefixName);
    seedAttr = DOM::QualifiedName(ATTR_SEED, DOM::emptyPrefixName);
    shape_renderingAttr = DOM::QualifiedName(ATTR_SHAPE_RENDERING, DOM::emptyPrefixName);
    slopeAttr = DOM::QualifiedName(ATTR_SLOPE, DOM::emptyPrefixName);
    spacingAttr = DOM::QualifiedName(ATTR_SPACING, DOM::emptyPrefixName);
    specularConstantAttr = DOM::QualifiedName(ATTR_SPECULARCONSTANT, DOM::emptyPrefixName);
    specularExponentAttr = DOM::QualifiedName(ATTR_SPECULAREXPONENT, DOM::emptyPrefixName);
    spreadMethodAttr = DOM::QualifiedName(ATTR_SPREADMETHOD, DOM::emptyPrefixName);
    startOffsetAttr = DOM::QualifiedName(ATTR_STARTOFFSET, DOM::emptyPrefixName);
    stdDeviationAttr = DOM::QualifiedName(ATTR_STDDEVIATION, DOM::emptyPrefixName);
    stemhAttr = DOM::QualifiedName(ATTR_STEMH, DOM::emptyPrefixName);
    stemvAttr = DOM::QualifiedName(ATTR_STEMV, DOM::emptyPrefixName);
    stitchTilesAttr = DOM::QualifiedName(ATTR_STITCHTILES, DOM::emptyPrefixName);
    stop_colorAttr = DOM::QualifiedName(ATTR_STOP_COLOR, DOM::emptyPrefixName);
    stop_opacityAttr = DOM::QualifiedName(ATTR_STOP_OPACITY, DOM::emptyPrefixName);
    strikethrough_positionAttr = DOM::QualifiedName(ATTR_STRIKETHROUGH_POSITION, DOM::emptyPrefixName);
    strikethrough_thicknessAttr = DOM::QualifiedName(ATTR_STRIKETHROUGH_THICKNESS, DOM::emptyPrefixName);
    strokeAttr = DOM::QualifiedName(ATTR_STROKE, DOM::emptyPrefixName);
    stroke_dasharrayAttr = DOM::QualifiedName(ATTR_STROKE_DASHARRAY, DOM::emptyPrefixName);
    stroke_dashoffsetAttr = DOM::QualifiedName(ATTR_STROKE_DASHOFFSET, DOM::emptyPrefixName);
    stroke_linecapAttr = DOM::QualifiedName(ATTR_STROKE_LINECAP, DOM::emptyPrefixName);
    stroke_linejoinAttr = DOM::QualifiedName(ATTR_STROKE_LINEJOIN, DOM::emptyPrefixName);
    stroke_miterlimitAttr = DOM::QualifiedName(ATTR_STROKE_MITERLIMIT, DOM::emptyPrefixName);
    stroke_opacityAttr = DOM::QualifiedName(ATTR_STROKE_OPACITY, DOM::emptyPrefixName);
    stroke_widthAttr = DOM::QualifiedName(ATTR_STROKE_WIDTH, DOM::emptyPrefixName);
    styleAttr = DOM::QualifiedName(ATTR_STYLE, DOM::emptyPrefixName);
    surfaceScaleAttr = DOM::QualifiedName(ATTR_SURFACESCALE, DOM::emptyPrefixName);
    systemLanguageAttr = DOM::QualifiedName(ATTR_SYSTEMLANGUAGE, DOM::emptyPrefixName);
    tableValuesAttr = DOM::QualifiedName(ATTR_TABLEVALUES, DOM::emptyPrefixName);
    targetAttr = DOM::QualifiedName(ATTR_TARGET, DOM::emptyPrefixName);
    targetXAttr = DOM::QualifiedName(ATTR_TARGETX, DOM::emptyPrefixName);
    targetYAttr = DOM::QualifiedName(ATTR_TARGETY, DOM::emptyPrefixName);
    text_anchorAttr = DOM::QualifiedName(ATTR_TEXT_ANCHOR, DOM::emptyPrefixName);
    text_decorationAttr = DOM::QualifiedName(ATTR_TEXT_DECORATION, DOM::emptyPrefixName);
    text_renderingAttr = DOM::QualifiedName(ATTR_TEXT_RENDERING, DOM::emptyPrefixName);
    textLengthAttr = DOM::QualifiedName(ATTR_TEXTLENGTH, DOM::emptyPrefixName);
    titleAttr = DOM::QualifiedName(ATTR_TITLE, DOM::emptyPrefixName);
    toAttr = DOM::QualifiedName(ATTR_TO, DOM::emptyPrefixName);
    transformAttr = DOM::QualifiedName(ATTR_TRANSFORM, DOM::emptyPrefixName);
    typeAttr = DOM::QualifiedName(ATTR_TYPE, DOM::emptyPrefixName);
    u1Attr = DOM::QualifiedName(ATTR_U1, DOM::emptyPrefixName);
    u2Attr = DOM::QualifiedName(ATTR_U2, DOM::emptyPrefixName);
    underline_positionAttr = DOM::QualifiedName(ATTR_UNDERLINE_POSITION, DOM::emptyPrefixName);
    underline_thicknessAttr = DOM::QualifiedName(ATTR_UNDERLINE_THICKNESS, DOM::emptyPrefixName);
    unicodeAttr = DOM::QualifiedName(ATTR_UNICODE, DOM::emptyPrefixName);
    unicode_bidiAttr = DOM::QualifiedName(ATTR_UNICODE_BIDI, DOM::emptyPrefixName);
    unicode_rangeAttr = DOM::QualifiedName(ATTR_UNICODE_RANGE, DOM::emptyPrefixName);
    units_per_emAttr = DOM::QualifiedName(ATTR_UNITS_PER_EM, DOM::emptyPrefixName);
    v_alphabeticAttr = DOM::QualifiedName(ATTR_V_ALPHABETIC, DOM::emptyPrefixName);
    v_hangingAttr = DOM::QualifiedName(ATTR_V_HANGING, DOM::emptyPrefixName);
    v_ideographicAttr = DOM::QualifiedName(ATTR_V_IDEOGRAPHIC, DOM::emptyPrefixName);
    v_mathematicalAttr = DOM::QualifiedName(ATTR_V_MATHEMATICAL, DOM::emptyPrefixName);
    valuesAttr = DOM::QualifiedName(ATTR_VALUES, DOM::emptyPrefixName);
    versionAttr = DOM::QualifiedName(ATTR_VERSION, DOM::emptyPrefixName);
    vert_adv_yAttr = DOM::QualifiedName(ATTR_VERT_ADV_Y, DOM::emptyPrefixName);
    vert_origin_xAttr = DOM::QualifiedName(ATTR_VERT_ORIGIN_X, DOM::emptyPrefixName);
    vert_origin_yAttr = DOM::QualifiedName(ATTR_VERT_ORIGIN_Y, DOM::emptyPrefixName);
    viewBoxAttr = DOM::QualifiedName(ATTR_VIEWBOX, DOM::emptyPrefixName);
    viewTargetAttr = DOM::QualifiedName(ATTR_VIEWTARGET, DOM::emptyPrefixName);
    visibilityAttr = DOM::QualifiedName(ATTR_VISIBILITY, DOM::emptyPrefixName);
    widthAttr = DOM::QualifiedName(ATTR_WIDTH, DOM::emptyPrefixName);
    widthsAttr = DOM::QualifiedName(ATTR_WIDTHS, DOM::emptyPrefixName);
    word_spacingAttr = DOM::QualifiedName(ATTR_WORD_SPACING, DOM::emptyPrefixName);
    writing_modeAttr = DOM::QualifiedName(ATTR_WRITING_MODE, DOM::emptyPrefixName);
    xAttr = DOM::QualifiedName(ATTR_X, DOM::emptyPrefixName);
    x_heightAttr = DOM::QualifiedName(ATTR_X_HEIGHT, DOM::emptyPrefixName);
    x1Attr = DOM::QualifiedName(ATTR_X1, DOM::emptyPrefixName);
    x2Attr = DOM::QualifiedName(ATTR_X2, DOM::emptyPrefixName);
    xChannelSelectorAttr = DOM::QualifiedName(ATTR_XCHANNELSELECTOR, DOM::emptyPrefixName);
    yAttr = DOM::QualifiedName(ATTR_Y, DOM::emptyPrefixName);
    y1Attr = DOM::QualifiedName(ATTR_Y1, DOM::emptyPrefixName);
    y2Attr = DOM::QualifiedName(ATTR_Y2, DOM::emptyPrefixName);
    yChannelSelectorAttr = DOM::QualifiedName(ATTR_YCHANNELSELECTOR, DOM::emptyPrefixName);
    zAttr = DOM::QualifiedName(ATTR_Z, DOM::emptyPrefixName);
    zoomAndPanAttr = DOM::QualifiedName(ATTR_ZOOMANDPAN, DOM::emptyPrefixName);
    WebCore::HTMLNames::classAttr = DOM::QualifiedName(ATTR_CLASS, DOM::emptyPrefixName);
    WebCore::XLinkNames::actuateAttr = DOM::QualifiedName(ATTR_XLINK_ACTUATE, DOM::emptyPrefixName);
    WebCore::XLinkNames::arcroleAttr = DOM::QualifiedName(ATTR_XLINK_ARCROLE, DOM::emptyPrefixName);
    WebCore::XLinkNames::hrefAttr = DOM::QualifiedName(ATTR_XLINK_HREF, DOM::emptyPrefixName);
    WebCore::XLinkNames::roleAttr = DOM::QualifiedName(ATTR_XLINK_ROLE, DOM::emptyPrefixName);
    WebCore::XLinkNames::showAttr = DOM::QualifiedName(ATTR_XLINK_SHOW, DOM::emptyPrefixName);
    WebCore::XLinkNames::titleAttr = DOM::QualifiedName(ATTR_XLINK_TITLE, DOM::emptyPrefixName);
    WebCore::XLinkNames::typeAttr = DOM::QualifiedName(ATTR_XLINK_TYPE, DOM::emptyPrefixName);

}
}
}

