#ifndef XLinkNames_H
#define XLinkNames_H

#include "misc/htmlnames.h"
#include "dom/QualifiedName.h"
#include "xml/Document.h"

namespace WebCore
{
namespace XLinkNames
{
extern DOM::QualifiedName actuateAttr;
extern DOM::QualifiedName arcroleAttr;
extern DOM::QualifiedName hrefAttr;
extern DOM::QualifiedName roleAttr;
extern DOM::QualifiedName showAttr;
extern DOM::QualifiedName titleAttr;
extern DOM::QualifiedName typeAttr;
}
}

#endif
